from .. import cdshealpix # noqa

import astropy.units as u
from astropy.coordinates import SkyCoord, Angle, Longitude, Latitude
import numpy as np

# Do not fill by hand :)
# > egrep "^ *def" healpix.py | cut -c 5- | egrep -v '^_'| cut -d '(' -f 1 | sed -r "s/^(.*)$/ '\1'/" | tr '\n' ','
__all__ = ['lonlat_to_healpix', 'skycoord_to_healpix', 'healpix_to_lonlat', 'healpix_to_skycoord', 'healpix_to_xy', 'vertices', 'vertices_skycoord']

# Raise a ValueError exception if the input 
# HEALPix cells array contains invalid values
def _check_ipixels(data, nside):
    npix = 12 * (nside ** 2)
    if (data >= npix).any() or (data < 0).any():
        valid_ipix = np.stack((np.zeros(npix.shape), npix)).T
        raise ValueError("The input HEALPix array contains values out of {0}.".format(valid_ipix))

def lonlat_to_healpix(lon, lat, nside, return_offsets=False, num_threads=0):
    r"""Get the HEALPix indexes that contains specific sky coordinates

    The ``nside`` of the returned HEALPix cell indexes must be specified. This 
    method is wrapped around the `hash <https://docs.rs/cdshealpix/0.1.5/cdshealpix/nested/struct.Layer.html#method.hash>`__ 
    method from the `cdshealpix Rust crate <https://crates.io/crates/cdshealpix>`__.

    Parameters
    ----------
    lon : `astropy.coordinates.Longitude`
        The longitudes of the sky coordinates.
    lat : `astropy.coordinates.Latitude`
        The latitudes of the sky coordinates.
    nside : `numpy.ndarray`
        The nside of the returned HEALPix cell indexes.
    return_offsets : bool, optional
        If set to `True`, returns a tuple made of 3 elements, the HEALPix cell
        indexes and the dx, dy arrays telling where the (``lon``, ``lat``) coordinates
        passed are located on the cells. ``dx`` and ``dy`` are :math:`\in [0, 1]`
    num_threads : int, optional
        Specifies the number of threads to use for the computation. Default to 0 means
        it will choose the number of threads based on the RAYON_NUM_THREADS environment variable (if set),
        or the number of logical CPUs (otherwise)

    Returns
    -------
    ipix : `numpy.ndarray`
        A numpy array containing all the HEALPix cell indexes stored as `np.uint64`.

    Raises
    ------
    ValueError
        When the number of longitudes and latitudes given do not match.

    Examples
    --------
    >>> from cdshealpix.ring import lonlat_to_healpix
    >>> from astropy.coordinates import Longitude, Latitude
    >>> import astropy.units as u
    >>> import numpy as np
    >>> lon = Longitude([0, 50, 25], u.deg)
    >>> lat = Latitude([6, -12, 45], u.deg)
    >>> depth = np.array([12, 14])
    >>> nside = 2 ** depth
    >>> ipix = lonlat_to_healpix(lon[:, np.newaxis], lat[:, np.newaxis], nside[np.newaxis, :])
    """
    # Check arrays
    #
    # We could have continued to use `.to_value(u.rad)` instead of `.rad`.
    # Although `to_value` is more generical (method of Quantity),
    # Longitude/Latitude ensure that the values the contain are in the correct ranges.
    lon = np.atleast_1d(lon.rad)
    lat = np.atleast_1d(lat.rad)
    nside = np.atleast_1d(nside)

    if (nside < 1).any() or (nside > (1 << 29)).any():
        raise ValueError("nside must be in the [1, (1 << 29)[ closed range")

    if lon.shape != lat.shape:
        raise ValueError("The number of longitudes does not match with the number of latitudes given")

    # Broadcasting
    lon, lat, nside = np.broadcast_arrays(lon, lat, nside)

    # Allocation of the array containing the resulting coordinates    
    num_ipix = lon.shape
    ipix = np.empty(num_ipix, dtype=np.uint64)
    dx = np.empty(num_ipix, dtype=np.float64)
    dy = np.empty(num_ipix, dtype=np.float64)

    # Call the Rust extension
    nside = nside.astype(np.uint32)
    num_threads = np.uint16(num_threads)
    cdshealpix.lonlat_to_healpix_ring(nside, lon, lat, ipix, dx, dy, num_threads)

    if return_offsets:
        return ipix, dx, dy
    else:
        return ipix

def skycoord_to_healpix(skycoord, nside, return_offsets=False, num_threads=0):
    r"""Get the HEALPix indexes that contains specific sky coordinates

    The ``nside`` of the returned HEALPix cell indexes must be specified.
    This method is wrapped around the
    `hash <https://docs.rs/cdshealpix/0.1.5/cdshealpix/nested/struct.Layer.html#method.hash>`__
    method from the `cdshealpix Rust crate <https://crates.io/crates/cdshealpix>`__.

    Parameters
    ----------
    skycoord : `astropy.coordinates.SkyCoord`
        The sky coordinates.
    nside : `numpy.ndarray`
        The nside of the returned HEALPix cell indexes.
    return_offsets : bool, optional
        If set to `True`, returns a tuple made of 3 elements, the HEALPix cell
        indexes and the dx, dy arrays telling where the (``lon``, ``lat``) coordinates
        passed are located in the cells. ``dx`` and ``dy`` are :math:`\in [0, 1]`
    num_threads : int, optional
        Specifies the number of threads to use for the computation. Default to 0 means
        it will choose the number of threads based on the RAYON_NUM_THREADS environment variable (if set),
        or the number of logical CPUs (otherwise)

    Returns
    -------
    ipix : `numpy.ndarray`
        A numpy array containing all the HEALPix cell indexes stored as `np.uint64`.

    Raises
    ------
    ValueError
        When the number of longitudes and latitudes given do not match.

    Examples
    --------
    >>> from cdshealpix.ring import skycoord_to_healpix
    >>> import astropy.units as u
    >>> from astropy.coordinates import SkyCoord
    >>> import numpy as np
    >>> skycoord = SkyCoord([0, 50, 25] * u.deg, [6, -12, 45] * u.deg, frame="icrs")
    >>> depth = 12
    >>> ipix = skycoord_to_healpix(skycoord, 1 << depth)
    """
    return lonlat_to_healpix(Longitude(skycoord.icrs.ra), Latitude(skycoord.icrs.dec), nside, return_offsets, num_threads)

def healpix_to_lonlat(ipix, nside, dx=0.5, dy=0.5, num_threads=0):
    r"""Get the longitudes and latitudes of the center of some HEALPix cells at a given depth.

    This method does the opposite transformation of `lonlat_to_healpix`.
    It's wrapped around the `center <https://docs.rs/cdshealpix/0.1.5/cdshealpix/nested/struct.Layer.html#method.center>`__
    method from the `cdshealpix Rust crate <https://crates.io/crates/cdshealpix>`__.

    Parameters
    ----------
    ipix : `numpy.ndarray`
        The HEALPix cell indexes given as a `np.uint64` numpy array.
    nside : `numpy.ndarray`
        The nside of the HEALPix cells.
    dx : float, optional
        The offset position :math:`\in [0, 1]` along the X axis. By default, `dx=0.5`
    dy : float, optional
        The offset position :math:`\in [0, 1]` along the Y axis. By default, `dy=0.5`
    num_threads : int, optional
        Specifies the number of threads to use for the computation. Default to 0 means
        it will choose the number of threads based on the RAYON_NUM_THREADS environment variable (if set),
        or the number of logical CPUs (otherwise)

    Returns
    -------
    lon, lat : (`astropy.coordinates.Longitude`, `astropy.coordinates.Latitude`)
        The sky coordinates of the center of the HEALPix cells given as a longitude, latitude tuple.

    Raises
    ------
    ValueError
        When the HEALPix cell indexes given have values out of :math:`[0, 12` x :math:`N_{side} ^ 2[`.

    Examples
    --------
    >>> from cdshealpix.ring import healpix_to_lonlat
    >>> import numpy as np
    >>> ipix = np.array([42, 6, 10], dtype=np.uint64)
    >>> depth = np.array([2, 12])
    >>> nside = 2 ** depth
    >>> lon, lat = healpix_to_lonlat(ipix[:, np.newaxis], nside[np.newaxis, :])
    """
    # Check arrays
    ipix = np.atleast_1d(ipix)
    nside = np.atleast_1d(nside)
    if (nside < 1).any() or (nside > (1 << 29)).any():
        raise ValueError("nside must be in the [1, (1 << 29)[ closed range")

    if dx < 0 or dx > 1:
        raise ValueError("dx must be between [0, 1]")

    if dy < 0 or dy > 1:
        raise ValueError("dy must be between [0, 1]")

    _check_ipixels(data=ipix, nside=nside)

    # Broadcasting
    ipix, nside = np.broadcast_arrays(ipix, nside)

    size_skycoords = ipix.shape
    # Allocation of the array containing the resulting coordinates
    lon = np.zeros(size_skycoords)
    lat = np.zeros(size_skycoords)

    # Call the Rust extension
    nside = nside.astype(np.uint32)
    ipix = ipix.astype(np.uint64)
    num_threads = np.uint16(num_threads)

    cdshealpix.healpix_to_lonlat_ring(nside, ipix, dx, dy, lon, lat, num_threads)
    return Longitude(lon, u.rad), Latitude(lat, u.rad)

def healpix_to_skycoord(ipix, nside, dx=0.5, dy=0.5, num_threads=0):
    r"""Get the sky coordinates of the center of some HEALPix cells at a given nside.

    This method does the opposite transformation of `lonlat_to_healpix`.
    It is the equivalent of `healpix_to_lonlat` except that it returns `astropy.coordinates.SkyCoord` instead
    of `astropy.units.Quantity`.
    It's wrapped around the `center <https://docs.rs/cdshealpix/0.1.5/cdshealpix/nested/struct.Layer.html#method.center>`__
    method from the `cdshealpix Rust crate <https://crates.io/crates/cdshealpix>`__.

    Parameters
    ----------
    ipix : `numpy.ndarray`
        The HEALPix cell indexes given as a `np.uint64` numpy array.
    nside : `numpy.ndarray`
        The nside of the HEALPix cells.
    dx : float, optional
        The offset position :math:`\in [0, 1]` along the X axis. By default, `dx=0.5`
    dy : float, optional
        The offset position :math:`\in [0, 1]` along the Y axis. By default, `dy=0.5`
    num_threads : int, optional
        Specifies the number of threads to use for the computation. Default to 0 means
        it will choose the number of threads based on the RAYON_NUM_THREADS environment variable (if set),
        or the number of logical CPUs (otherwise)

    Returns
    -------
    skycoord : `astropy.coordinates.SkyCoord`
        The sky coordinates of the center of the HEALPix cells given as a `~astropy.coordinates.SkyCoord` object.

    Raises
    ------
    ValueError
        When the HEALPix cell indexes given have values out of :math:`[0, 12` x :math:`N_{side} ^ 2[`.

    Examples
    --------
    >>> from cdshealpix.ring import healpix_to_skycoord
    >>> import numpy as np
    >>> ipix = np.array([42, 6, 10])
    >>> depth = 12
    >>> nside = 2 ** depth
    >>> skycoord = healpix_to_skycoord(ipix, nside)
    """
    lon, lat = healpix_to_lonlat(ipix, nside, dx, dy, num_threads)
    return SkyCoord(ra=lon, dec=lat, frame="icrs", unit="rad")

def healpix_to_xy(ipix, nside, num_threads=0):
    r"""
    Project the center of a HEALPix cell to the xy-HEALPix plane

    Parameters
    ----------
    ipix : `numpy.ndarray`
        The HEALPix cells which centers will be projected
    nside : `numpy.ndarray`
        The nside of the HEALPix cells
    num_threads : int, optional
        Specifies the number of threads to use for the computation. Default to 0 means
        it will choose the number of threads based on the RAYON_NUM_THREADS environment variable (if set),
        or the number of logical CPUs (otherwise)

    Returns
    -------
    x, y: (`numpy.ndarray`, `numpy.ndarray`)
        The position of the HEALPix centers in the xy-HEALPix plane.
        :math:`x \in [0, 8[` and :math:`y \in [-2, 2]`

    Examples
    --------
    >>> from cdshealpix.ring import healpix_to_xy
    >>> import astropy.units as u
    >>> import numpy as np
    >>> depth = np.array([0, 10])
    >>> nside = 2 ** depth
    >>> ipix = np.arange(12)
    >>> x, y = healpix_to_xy(ipix[:, np.newaxis], nside[np.newaxis, :])
    """
    # Check arrays
    ipix = np.atleast_1d(ipix)
    nside = np.atleast_1d(nside)

    if (nside < 1).any() or (nside > (1 << 29)).any():
        raise ValueError("nside must be in the [1, (1 << 29)[ closed range")

    _check_ipixels(data=ipix, nside=nside)

    # Broadcasting
    ipix, nside = np.broadcast_arrays(ipix, nside)

    # Allocation for the resulting arrays
    x = np.zeros(ipix.shape, dtype=np.float64)
    y = np.zeros(ipix.shape, dtype=np.float64)

    # Call the Rust extension
    ipix = ipix.astype(np.uint64)
    nside = nside.astype(np.uint32)
    num_threads = np.uint16(num_threads)
    cdshealpix.healpix_to_xy_ring(nside, ipix, x, y, num_threads)

    return x, y

def vertices(ipix, nside, step=1, num_threads=0):
    """Get the longitudes and latitudes of the vertices of some HEALPix cells at a given nside.

    This method returns the 4 vertices of each cell in `ipix`.
    This method is wrapped around the `vertices <https://docs.rs/cdshealpix/0.1.5/cdshealpix/nested/struct.Layer.html#method.vertices>`__
    method from the `cdshealpix Rust crate <https://crates.io/crates/cdshealpix>`__.

    Parameters
    ----------
    ipix : `numpy.ndarray`
        The HEALPix cell indexes given as a `np.uint64` numpy array.
    nside : int
        The nside of the HEALPix cells.
    step : int, optional
        The number of vertices returned per HEALPix side. By default it is set to 1 meaning that
        it will only return the vertices of the cell. 2 means that it will returns the vertices of
        the cell plus one more vertex per edge (the middle of it). More generally, the number
        of vertices returned is ``4 * step``.
    num_threads : int, optional
        Specifies the number of threads to use for the computation. Default to 0 means
        it will choose the number of threads based on the RAYON_NUM_THREADS environment variable (if set),
        or the number of logical CPUs (otherwise)

    Returns
    -------
    lon, lat : (`astropy.coordinates.Longitude`, `astropy.coordinates.Latitude`)
        The sky coordinates of the 4 vertices of the HEALPix cells. 
        `lon` and `lat` are `~astropy.coordinates.Longitude` and `~astropy.coordinates.Latitude` instances respectively, 
        containing a :math:`N` x :math:`4` numpy array where N is the number of HEALPix cell given in `ipix`.

    Warnings
    --------
    ``step`` is currently not implemented for the ring scheme. Therefore it is set to 1 by default.

    Raises
    ------
    ValueError
        When the HEALPix cell indexes given have values out of :math:`[0, 12` x :math:`N_{side} ^ 2[`.

    Examples
    --------
    >>> from cdshealpix.ring import vertices
    >>> import numpy as np
    >>> ipix = np.array([42, 6, 10])
    >>> depth = 12
    >>> lon, lat = vertices(ipix, (1 << depth))
    """
    if nside < 1 or nside > (1 << 29):
        raise ValueError("nside must be in the [1, (1 << 29)[ closed range")

    if step < 1:
        raise ValueError("The number of step must be >= 1")

    ipix = np.atleast_1d(ipix)
    _check_ipixels(data=ipix, nside=nside)
    ipix = ipix.astype(np.uint64)

    # Allocation of the array containing the resulting coordinates
    lon = np.zeros(ipix.shape + (4 * step,))
    lat = np.zeros(ipix.shape + (4 * step,))
    num_threads = np.uint16(num_threads)

    cdshealpix.vertices_ring(nside, ipix, step, lon, lat, num_threads)
    return Longitude(lon, u.rad), Latitude(lat, u.rad)

def vertices_skycoord(ipix, nside, step=1):
    """Get the sky coordinates of the vertices of some HEALPix cells at a given nside.

    This method returns the 4 vertices of each cell in `ipix`.
    This method is wrapped around the `vertices <https://docs.rs/cdshealpix/0.1.5/cdshealpix/nested/struct.Layer.html#method.vertices>`__
    method from the `cdshealpix Rust crate <https://crates.io/crates/cdshealpix>`__.

    Parameters
    ----------
    ipix : `numpy.ndarray`
        The HEALPix cell indexes given as a `np.uint64` numpy array.
    nside : int
        The nside of the HEALPix cells.
    step : int, optional
        The number of vertices returned per HEALPix side. By default it is set to 1 meaning that
        it will only return the vertices of the cell. 2 means that it will returns the vertices of
        the cell plus one more vertex per edge (the middle of it). More generally, the number
        of vertices returned is ``4 * step``.

    Returns
    -------
    vertices : `astropy.coordinates.SkyCoord`
        The sky coordinates of the 4 vertices of the HEALPix cells. `vertices` is a `~astropy.coordinates.SkyCoord` object
        containing a :math:`N` x :math:`4` numpy array where N is the number of HEALPix cells given in `ipix`.

    Raises
    ------
    ValueError
        When the HEALPix cell indexes given have values out of :math:`[0, 12` x :math:`N_{side} ^ 2[`.

    Examples
    --------
    >>> from cdshealpix.ring import vertices_skycoord
    >>> import numpy as np
    >>> ipix = np.array([42, 6, 10])
    >>> depth = 12
    >>> vertices = vertices_skycoord(ipix, 1 << depth)
    """
    lon, lat = vertices(ipix, nside, step)
    return SkyCoord(ra=lon, dec=lat, frame="icrs", unit="rad")



import pytest
import numpy as np

from astropy.coordinates import Angle, SkyCoord, Longitude, Latitude
import astropy.units as u

from ..nested import lonlat_to_healpix, \
 healpix_to_lonlat, \
 healpix_to_xy, \
 lonlat_to_xy, \
 skycoord_to_healpix, \
 xy_to_lonlat, \
 healpix_to_skycoord, \
 vertices, \
 neighbours, \
 cone_search, \
 polygon_search, \
 elliptical_cone_search, \
 external_neighbours, \
 bilinear_interpolation


from .. import to_ring, from_ring

@pytest.mark.parametrize("size", [1, 10, 100, 1000, 10000, 100000])
def test_lonlat_to_healpix(size):
    depth = np.random.randint(30)
    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    ipixels, dx, dy = lonlat_to_healpix(lon=lon, lat=lat, depth=depth, return_offsets=True)
    ipixels, dx, dy = skycoord_to_healpix(
        SkyCoord(lon, lat, frame="icrs"),
        depth=depth,
        return_offsets=True,
    )

    npix = 12 * 4**(depth)
    assert ((ipixels >= 0) & (ipixels < npix)).all()
    assert ((dx >= 0) & (dx <= 1)).all()

@pytest.mark.parametrize("size", [1, 10, 100, 1000, 10000, 100000])
def test_healpix_to_lonlat(size):
    depth = np.random.randint(30)
    size = 10000
    ipixels = np.random.randint(12 * 4 ** depth, size=size, dtype=np.uint64)

    lon, lat = healpix_to_lonlat(ipix=ipixels, depth=depth)
    assert lon.shape == lat.shape

def test_healpix_to_lonlat_on_broadcasted_arrays():
    depth = 12
    x = np.arange(1000000)
    healpix_to_lonlat(ipix=x, depth=depth)
    y = x[::2]
    healpix_to_lonlat(ipix=y, depth=depth)

    z = np.array([1, 2, 3])
    a = np.broadcast_to(z, (10000, 3))
    healpix_to_lonlat(ipix=a, depth=depth)

    b = np.broadcast_to(3, (3, 4, 5))
    healpix_to_lonlat(ipix=b, depth=depth)

def test_healpix_to_lonlat_on_broadcasted_arrays2():
    level = np.arange(2)
    nside = 2**level

    ipix = np.arange(3)

    lon, lat = healpix_to_lonlat(ipix[:, np.newaxis], level[np.newaxis, :])
    assert lon.shape == lat.shape and lon.shape == (3, 2)

def test_invalid_depth_exception():
    size = 10000
    ipix = np.zeros(size)
    with pytest.raises(Exception):
        healpix_to_lonlat(ipix, -2)
    with pytest.raises(Exception):
        healpix_to_lonlat(ipix, 30)
    with pytest.raises(Exception):
        lonlat_to_healpix(Longitude(0, u.deg), Latitude(0, u.deg), -2)
    with pytest.raises(Exception):
        vertices(ipix, -2)
    with pytest.raises(Exception):
        neighbours(ipix, -2)
    with pytest.raises(Exception):
        cone_search(Longitude(0, u.deg), Latitude(0, u.deg), 15 * u.deg, -2)

def test_lonlat_shape_exception():
    lon = Longitude([2, 5], u.deg)
    lat = Latitude([5], u.deg)
    
    with pytest.raises(Exception):
        lonlat_to_healpix(lon, lat, 12)

@pytest.mark.parametrize("depth", [0, 12, 24])
def test_invalid_ipix_exception(depth):
    npix = 12 * 4 ** depth
    invalid_ipix1 = np.array([-20, 0, 11])
    invalid_ipix2 = np.array([0, npix + 1, 11])
    with pytest.raises(Exception):
        healpix_to_lonlat(invalid_ipix1, depth)
    with pytest.raises(Exception):
        healpix_to_lonlat(invalid_ipix2, depth)
    with pytest.raises(Exception):
        vertices(invalid_ipix1, depth)
    with pytest.raises(Exception):
        neighbours(invalid_ipix1, depth)

def test_healpix_to_skycoord():
    ipix = np.array([0, 2, 4])
    skycoord = healpix_to_skycoord(ipix=ipix, depth=0)
    assert skycoord.icrs.ra.shape == skycoord.icrs.dec.shape

def test_vertices_lonlat():
    depth = 12
    size = 100000
    ipixels = np.random.randint(12 * 4 ** depth, size=size)

    lon, lat = vertices(ipix=ipixels, depth=depth)
    assert lon.shape == lat.shape
    assert lon.shape == (size, 4)

def test_neighbours():
    depth = 0
    size = 100000
    ipixels = np.random.randint(12 * 4**(depth), size=size)

    neigh = neighbours(ipix=ipixels, depth=depth)
    assert neigh.shape == (size, 9)

    npix = 12 * 4**(depth)
    assert ((neigh >= -1) & (neigh < npix)).all()

def test_cone_search():
    lon = Longitude(np.random.rand(1)[0] * 360, u.deg)
    lat = Latitude(np.random.rand(1)[0] * 180 - 90, u.deg)
    radius = (np.random.rand(1)[0] * 45) * u.deg
    max_depth = 5

    ipix, depth, fully_covered = cone_search(lon=lon, lat=lat, radius=radius, depth=max_depth, flat=True)

    npix = 12 * 4 ** (max_depth)
    assert ((depth >= 0) & (depth <= max_depth)).all()
    assert ((ipix >= 0) & (ipix < npix)).all()

    with pytest.raises(Exception):
        cone_search(Longitude([5, 4], u.deg), Latitude([5, 4], u.deg), 15 * u.deg, depth=12)

@pytest.mark.parametrize("size", [0, 1, 2, 3, 5, 6, 9])
def test_polygon_search(size):
    max_depth = 12
    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    if size < 3:
        with pytest.raises(Exception):
            polygon_search(lon=lon, lat=lat, depth=max_depth)
    else:
        ipix, depth, fully_covered = polygon_search(lon=lon, lat=lat, depth=max_depth)

        npix = 12 * 4 ** (max_depth)
        assert ((depth >= 0) & (depth <= max_depth)).all()
        assert ((ipix >= 0) & (ipix < npix)).all()

# From https://github.com/cds-astro/cds-healpix-python/issues/10
def test_polygon_search_issue10():
    coords = SkyCoord([(353.8156714, -56.33202193), (6.1843286, -56.33202193), (5.27558041, -49.49378172), (354.72441959, -49.49378172)], unit=u.deg)
    polygon_search(Longitude(coords.ra), Latitude(coords.dec), 12)

# From https://github.com/cds-astro/mocpy/issues/57
def test_polygon_search_issue57():
    coords = np.load('./cdshealpix/tests/moc_coords.npy')
    polygon_search(Longitude(coords[:,0], u.deg), Latitude(coords[:,1], u.deg), 9)

def test_polygon_search_not_enough_vertices_exception():
    # 4 total vertices but only 2 distincts. This should fail
    with pytest.raises(Exception):
        polygon_search(Longitude([1, 1, 2, 1], u.deg), Latitude([1, 1, 3, 1], u.deg), depth=12)
    
    # 4 total vertices and 3 distincts. This should pass
    polygon_search(Longitude([1, 1, 2, 1], u.deg), Latitude([1, 1, 3, 2], u.deg), depth=12)

# Following an error spotted by the hips2fits service that usually deal with big polygons
def test_polygon_search_big_polygon():
    ipix, _, _ = polygon_search(
        Longitude([268.84102386, 299.40278164, 66.0951825,   96.66953469], u.deg),
        Latitude([-45.73283624, 38.47909742, 38.76894431, -45.43470273], u.deg),
        depth=0,
        flat=True
    )
    assert (ipix == np.asarray([0, 3, 4, 5, 7, 8, 9, 10, 11])).all()

def test_elliptical_cone_search():
    lon = Longitude(0, u.deg)
    lat = Latitude(0, u.deg)
    a = Angle(50, unit="deg")
    b = Angle(5, unit="deg")
    pa = Angle(0, unit="deg")
    max_depth = 12

    ipix, depth, fully_covered = elliptical_cone_search(lon, lat, a, b, pa, max_depth)

    npix = 12 * 4 ** (max_depth)
    assert ((depth >= 0) & (depth <= max_depth)).all()
    assert ((ipix >= 0) & (ipix < npix)).all()

@pytest.mark.parametrize("depth,ipix,expected_border_cells,expected_corner_cells", [
    (0, 0, np.array([90, 91, 94, 95, 26, 27, 30, 31, 53, 55, 61, 63, 69, 71, 77, 79]), np.array([143, -1, 47, -1])),
    (27, 0, np.array(
        [1633305464859699882, 1633305464859699883, 1633305464859699886, 1633305464859699887, 16, 18,
         24, 26, 32, 33, 36, 37, 1248998296657417557, 1248998296657417559, 1248998296657417565, 1248998296657417567]),
        np.array([2594073385365405695, 1633305464859699898, 48, 1248998296657417589]))
])
def test_external_neighbours(depth, ipix, expected_border_cells, expected_corner_cells):
    delta_depth = 2
    ipix_border_cells, ipix_corner_cells = external_neighbours(ipix, depth, delta_depth)
    assert (expected_border_cells == ipix_border_cells).all()
    assert (expected_corner_cells == ipix_corner_cells).all()

@pytest.mark.parametrize("ipix, depth, expected_x, expected_y", [
    (np.arange(12), 0,
     np.array([1., 3., 5., 7., 0., 2., 4., 6., 1., 3., 5., 7.], dtype=np.float64),
     np.array([1., 1., 1., 1., 0., 0., 0., 0., -1., -1., -1., -1.], dtype=np.float64)),
    ([], 0,
     np.array([], dtype=np.float64),
     np.array([], dtype=np.float64)),
    (0, 0,
     np.array([1.], dtype=np.float64),
     np.array([1.], dtype=np.float64))
])
def test_healpix_to_xy(ipix, depth, expected_x, expected_y):
    x, y = healpix_to_xy(ipix, depth)

    assert (x == expected_x).all()
    assert (y == expected_y).all()

def test_healpix_to_xy_expection():
    with pytest.raises(ValueError):
        healpix_to_xy(np.array([]), -5)
    
    with pytest.raises(ValueError):
        healpix_to_xy(np.array([-5]), 12)

def assert_equal_array(first, second, tol=1e-8):
    assert ((first - second) < tol).all()

@pytest.mark.parametrize("lon, lat, expected_x, expected_y", [
    (Longitude(np.array([0., 0.78539816, 1.57079633]), u.rad),
     Latitude(np.array([-0.72972766, 0., 0.72972766]), u.rad),
     np.array([0., 1., 2.]),
     np.array([-1., 0., 1.]))
])
def test_lonlat_to_xy(lon, lat, expected_x, expected_y):
    x, y = lonlat_to_xy(lon, lat)
    assert_equal_array(x, expected_x)
    assert_equal_array(y, expected_y)

@pytest.mark.parametrize("x, y, expected_lon, expected_lat", [
    (np.array([0., 1., 2.]),
     np.array([-1., 0., 1.]),
     np.array([0., 0.78539816, 1.57079633]) * u.rad,
     np.array([-0.72972766, 0., 0.72972766]) * u.rad)
])
def test_xy_to_lonlat(x, y, expected_lon, expected_lat):
    lon, lat = xy_to_lonlat(x, y)
    assert_equal_array(lon.to_value(u.rad), expected_lon.to_value(u.rad))
    assert_equal_array(lat.to_value(u.rad), expected_lat.to_value(u.rad))

### Test to_ring & from_ring
@pytest.mark.parametrize("pix, depth, expected_ring_pix", [
    (np.arange(12), 0, np.arange(12)),
    (4, 1, 15),
    (np.array([4, 5, 6, 7]), 1, np.array([15, 7, 6, 1])),
    (np.array([28, 29, 30, 31]), 1, np.array([34, 26, 25, 18]))
])
def test_to_ring(pix, depth, expected_ring_pix):
    ring_pix = to_ring(pix, depth)
    assert (ring_pix == expected_ring_pix).all()

@pytest.mark.parametrize("pix, depth, expected_nested_pix", [
    (np.arange(12), 0, np.arange(12)),
    (15, 1, 4),
    (np.array([15, 7, 6, 1]), 1, np.array([4, 5, 6, 7])),
    (np.array([34, 26, 25, 18]), 1, np.array([28, 29, 30, 31]))
])
def test_from_ring(pix, depth, expected_nested_pix):
    nested_pix = from_ring(pix, depth)
    assert (nested_pix == expected_nested_pix).all()

@pytest.mark.parametrize("size", [1, 10, 100, 1000, 10000, 100000])
def test_from_vs_to_ring(size):
    depth = np.random.randint(30)
    nside = 1 << depth

    ipixels = np.random.randint(12 * (nside ** 2), size=size, dtype="uint64")
    ring = to_ring(ipix=ipixels, depth=depth)
    ipixels_result = from_ring(ipix=ring, depth=depth)
    assert (ipixels == ipixels_result).all() 

@pytest.mark.parametrize("depth", [5, 0, 7, 12, 20, 29])
def test_bilinear_interpolation(depth):
    size = 1000

    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    ipix, weights = bilinear_interpolation(lon, lat, depth)
    
    assert ((weights >= 0.0) & (weights <= 1.0)).all()
    assert weights.sum() == ipix.shape[0]
    assert ((ipix >= 0) & (ipix < 12 * (4 ** depth))).all()

@pytest.mark.parametrize("depth", [5, 0, 7, 12, 20, 29])
def test_bilinear_interpolation(depth):
    size = 1000

    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    ipix, weights = bilinear_interpolation(lon, lat, depth)
    
    assert ((weights >= 0.0) & (weights <= 1.0)).all()
    assert weights.sum() == ipix.shape[0]
    assert ((ipix >= 0) & (ipix < 12 * (4 ** depth))).all()

@pytest.mark.parametrize("depth", [5, 0, 7, 12, 20, 29])
def test_bilinear_interpolation2(depth):
    lon = Longitude([10, 25, 0], u.deg)
    lat = Latitude([5, 10, 45], u.deg)
    depth = 5

    ipix, weights = bilinear_interpolation(lon, lat, depth)





import pytest
import astropy.units as u
import numpy as np
import astropy_healpix

from ..nested import lonlat_to_healpix, \
 healpix_to_lonlat, \
 healpix_to_skycoord, \
 vertices, \
 neighbours, \
 cone_search
from astropy.coordinates import Longitude, Latitude

@pytest.mark.benchmark(group="lonlat_to_healpix_broadcast")
def test_lonlat_to_healpix_broadcast(benchmark):
    size = 10000
    depth = np.random.randint(low=0, high=29, size=10)

    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    ipixels = benchmark(lonlat_to_healpix,
        lon=lon[:, np.newaxis],
        lat=lat[:, np.newaxis],
        depth=depth[np.newaxis, :],
        num_threads=1)

@pytest.mark.benchmark(group="lonlat_to_healpix_broadcast")
def test_lonlat_to_healpix_broadcast_multithread(benchmark):
    size = 10000
    depth = np.random.randint(low=0, high=29, size=10)

    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    ipixels = benchmark(lonlat_to_healpix,
        lon=lon[:, np.newaxis],
        lat=lat[:, np.newaxis],
        depth=depth[np.newaxis, :])



@pytest.mark.benchmark(group="lonlat_to_healpix_broadcast")
def test_lonlat_to_healpix_astropy_broadcast(benchmark):
    depth = np.random.randint(low=0, high=29, size=10)
    nside = 2 ** depth

    size = 10000
    lon = np.random.rand(size) * 360 * u.deg
    lat = (np.random.rand(size) * 180 - 90) * u.deg

    ipixels = benchmark(astropy_healpix.core.lonlat_to_healpix,
        lon=lon[:, np.newaxis],
        lat=lat[:, np.newaxis],
        nside=nside[np.newaxis, :],
        order='nested')

@pytest.mark.benchmark(group="lonlat_to_healpix")
def test_lonlat_to_healpix(benchmark):
    size = 10000
    depth = 12
    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    ipixels = benchmark(lonlat_to_healpix, lon=lon, lat=lat, depth=depth, num_threads=1)

@pytest.mark.benchmark(group="lonlat_to_healpix")
def test_lonlat_to_healpix_multithread(benchmark):
    size = 10000
    depth = 12
    lon = Longitude(np.random.rand(size) * 360, u.deg)
    lat = Latitude(np.random.rand(size) * 180 - 90, u.deg)

    ipixels = benchmark(lonlat_to_healpix, lon=lon, lat=lat, depth=depth)

@pytest.mark.benchmark(group="lonlat_to_healpix")
def test_lonlat_to_healpix_astropy(benchmark):
    size = 10000
    depth = 12
    nside = 1 << depth
    lon = np.random.rand(size) * 360 * u.deg
    lat = (np.random.rand(size) * 180 - 90) * u.deg

    ipixels = benchmark(astropy_healpix.core.lonlat_to_healpix, lon=lon, lat=lat, nside=nside, order='nested')

@pytest.mark.benchmark(group="healpix_to_lonlat_broadcast")
def test_healpix_to_lonlat_broadcast(benchmark):
    size = 1000
    depth = np.random.randint(low=0, high=29, size=10)
    ipixels = np.random.randint(12 * 4 ** np.min(depth), size=size)

    lon, lat = benchmark(healpix_to_lonlat, ipix=ipixels[:, np.newaxis], depth=depth[np.newaxis, :], num_threads=1)

@pytest.mark.benchmark(group="healpix_to_lonlat_broadcast")
def test_healpix_to_lonlat_broadcast_multithread(benchmark):
    size = 1000
    depth = np.random.randint(low=0, high=29, size=10)
    ipixels = np.random.randint(12 * 4 ** np.min(depth), size=size)

    lon, lat = benchmark(healpix_to_lonlat, ipix=ipixels[:, np.newaxis], depth=depth[np.newaxis, :])


@pytest.mark.benchmark(group="healpix_to_lonlat_broadcast")
def test_healpix_to_lonlat_astropy_broadcast(benchmark):
    size = 1000
    depth = np.random.randint(low=0, high=29, size=10)
    nside = 2 ** depth
    ipixels = np.random.randint(12 * 4 ** np.min(depth), size=size)

    lon, lat = benchmark(astropy_healpix.core.healpix_to_lonlat,
        healpix_index=ipixels[:, np.newaxis],
        nside=nside[np.newaxis, :],
        order='nested')

@pytest.mark.benchmark(group="healpix_to_lonlat")
def test_healpix_to_lonlat(benchmark):
    size = 10000
    depth = 12
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    lon, lat = benchmark(healpix_to_lonlat, ipix=ipixels, depth=depth, num_threads=1)

@pytest.mark.benchmark(group="healpix_to_lonlat")
def test_healpix_to_lonlat_multithread(benchmark):
    size = 10000
    depth = 12
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    lon, lat = benchmark(healpix_to_lonlat, ipix=ipixels, depth=depth)

@pytest.mark.benchmark(group="healpix_to_lonlat")
def test_healpix_to_lonlat_astropy(benchmark):
    size = 10000
    depth = 12
    nside = 1 << depth
    ipixels = np.random.randint(12 * 4 ** (depth), size=size)

    lon, lat = benchmark(astropy_healpix.core.healpix_to_lonlat,
        healpix_index=ipixels,
        nside=nside,
        order='nested')

def test_healpix_to_skycoord():
    skycoord = healpix_to_skycoord(ipix=[0, 2, 4], depth=0)

@pytest.mark.benchmark(group="vertices")
def test_healpix_vertices_lonlat(benchmark):
    depth = 12
    size = 100000
    ipixels = np.random.randint(12 * 4**(depth), size=size)

    lon, lat = benchmark(vertices, ipix=ipixels, depth=depth, num_threads=1)

@pytest.mark.benchmark(group="vertices")
def test_healpix_vertices_lonlat_multithread(benchmark):
    depth = 12
    size = 100000
    ipixels = np.random.randint(12 * 4**(depth), size=size)

    lon, lat = benchmark(vertices, ipix=ipixels, depth=depth)

@pytest.mark.benchmark(group="vertices")
def test_healpix_vertices_lonlat_astropy(benchmark):
    depth = 12
    size = 100000
    ipixels = np.random.randint(12 * 4**(depth), size=size)

    lon2, lat2 = benchmark(astropy_healpix.core.boundaries_lonlat, healpix_index=ipixels, nside=(1 << depth), step=1, order='nested')

@pytest.mark.benchmark(group="neighbours")
def test_healpix_neighbours(benchmark):
    depth = 12
    size = 100000
    ipixels = np.random.randint(12 * 4**(depth), size=size)

    res = benchmark(neighbours, ipix=ipixels, depth=depth)

lon_c = np.random.rand(1)[0] * 360 * u.deg
lat_c = (np.random.rand(1)[0] * 180 - 90) * u.deg
radius_c = np.random.rand(1)[0] * 45 * u.deg
depth_c = 5

@pytest.mark.benchmark(group="cone_search")
def test_cone_search(benchmark):
    res = benchmark.pedantic(cone_search, kwargs={'lon': Longitude(lon_c), 'lat': Latitude(lat_c), 'radius': radius_c, 'depth': depth_c}, iterations=10, rounds=100)

@pytest.mark.benchmark(group="cone_search")
def test_cone_search_astropy(benchmark):
    nside = 1 << depth_c
    res = benchmark.pedantic(astropy_healpix.core.healpix_cone_search, kwargs={'lon': lon_c, 'lat': lat_c, 'radius': radius_c, 'nside': nside, 'order': 'nested'}, iterations=10, rounds=100)



import numpy as np

from . import cdshealpix # noqa

__all__ = ['to_ring', 'from_ring']

# Raise a ValueError exception if the input 
# HEALPix cells array contains invalid values
def _check_ipixels(data, depth):
    npix = 12 * 4 ** (depth)
    if (data >= npix).any() or (data < 0).any():
        raise ValueError("The input HEALPix cells contains value out of [0, {0}]".format(npix - 1))


def to_ring(ipix, depth, num_threads=0):
    """Convert HEALPix cells from the NESTED to the RING scheme

    Parameters
    ----------
    ipix : `numpy.ndarray`
        The HEALPix cell indexes in the NESTED scheme.
    depth : int
        The depth of the HEALPix cells.
    num_threads : int, optional
        Specifies the number of threads to use for the computation. Default to 0 means
        it will choose the number of threads based on the RAYON_NUM_THREADS environment variable (if set),
        or the number of logical CPUs (otherwise)

    Returns
    -------
    ipix_ring : `numpy.ndarray`
        The corresponding HEALPix cells in the RING scheme.

    Raises
    ------
    ValueError
        When the HEALPix cell indexes given have values out of :math:`[0, 4^{29 - depth}[`.

    Examples
    --------
    >>> from cdshealpix import to_ring
    >>> import numpy as np
    >>> ipix = np.array([42, 6, 10])
    >>> depth = 12
    >>> ipix_ring = to_ring(ipix, depth)
    """
    if depth < 0 or depth > 29:
        raise ValueError("Depth must be in the [0, 29] closed range")

    ipix = np.atleast_1d(ipix)
    _check_ipixels(data=ipix, depth=depth)
    ipix = ipix.astype(np.uint64)
    
    # Allocation of the array containing the cells under the RING scheme
    ipix_ring = np.zeros(ipix.shape, dtype=np.uint64)

    num_threads = np.uint16(num_threads)
    cdshealpix.to_ring(depth, ipix, ipix_ring, num_threads)

    return ipix_ring

def from_ring(ipix, depth, num_threads=0):
    """Convert HEALPix cells from the RING to the NESTED scheme

    Parameters
    ----------
    ipix : `numpy.ndarray`
        The HEALPix cell indexes in the RING scheme.
    depth : int
        The depth of the HEALPix cells.
    num_threads : int, optional
        Specifies the number of threads to use for the computation. Default to 0 means
        it will choose the number of threads based on the RAYON_NUM_THREADS environment variable (if set),
        or the number of logical CPUs (otherwise)

    Returns
    -------
    ipix_nested : `numpy.ndarray`
        The corresponding HEALPix cells in the NESTED scheme.

    Raises
    ------
    ValueError
        When the HEALPix cell indexes given have values out of :math:`[0, 4^{29 - depth}[`.

    Examples
    --------
    >>> from cdshealpix import from_ring
    >>> import numpy as np
    >>> ipix = np.array([42, 6, 10])
    >>> depth = 12
    >>> ipix = from_ring(ipix, depth)
    """
    if depth < 0 or depth > 29:
        raise ValueError("Depth must be in the [0, 29] closed range")

    ipix = np.atleast_1d(ipix)
    _check_ipixels(data=ipix, depth=depth)
    ipix = ipix.astype(np.uint64)
    
    # Allocation of the array containing the cells under the NESTED scheme
    ipix_nested = np.zeros(ipix.shape, dtype=np.uint64)

    num_threads = np.uint16(num_threads)
    cdshealpix.from_ring(depth, ipix, ipix_nested, num_threads)

    return ipix_nested

